/*
+----------------------------------------------------------------------+
|                    Class: CLauncher                                  |
|                                                                      |
| Developper:  Eric Gavaldo (eric.gavaldo@xqual.com)                   |
| Version:     1.0                                                     |
| License:     GPL v2 (https://www.gnu.org/licenses/gpl-2.0.html)      |
+----------------------------------------------------------------------+
*/
package com.xqual.xlauncher.mrsuricate;

import java.io.File;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JOptionPane;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.xqual.xagent.CScannedTest;
import com.xqual.xagent.CTestExecution;
import com.xqual.xagent.launcher.CExecutionStep;
import com.xqual.xagent.launcher.CLauncher;
import com.xqual.xagent.launcher.CReturnStatus;
import com.xqual.xcommon.IConstantsResults;
import com.xqual.xcommon.image.CImageUtils;
import com.xqual.xcommon.type.CParamParsingException;
import com.xqual.xcommon.type.CTestAttribute;
import com.xqual.xcommon.type.CTestcaseParam;
import com.xqual.xcommon.utils.CCalendarUtils;
import com.xqual.xcommon.utils.CFileUtils;
import com.xqual.xcommon.utils.CUtils;



/**
 * The <code>CLauncherImpl</code> implementation of <code>ILauncher</code> for Mr Suricate.
 * @author egavaldo
 */
public class CLauncherImpl extends CLauncher implements IConstantsResults {

	// +==============================================================+
	// | Attributes                                                   |
	// +==============================================================+

	static final String TRACE_HEADER = "{mrsuricate    }  ";

	private String restBaseUrl;
	private String restEmail;
	private String restPassword;
	private String customerId;
	private int intCustomerId;
	private String browser;

	private CLauncherRestBackend restBackend;
	
	private final Hashtable<String, Long> hashtableScenarioNameId = new Hashtable<String, Long>();
	private final Hashtable<Long, Long>   hashtableScenarioRunIds = new Hashtable<Long, Long>();
	
	// +==============================================================+
	// | Constructors                                                 |
	// +==============================================================+

	public CLauncherImpl() {
		super(TRACE_HEADER);
	}

	// +==============================================================+
	// | Methods                                                      |
	// +==============================================================+

	@Override
	public CReturnStatus initialize(int sutId, String sutName, String sutVersion) { 
		setSutDetails(sutId, sutName, sutVersion);

		System.out.println("- - - - - - - - - - - - - - - - - - - - - - - - ");
		System.out.println("- - - - - - - - - - Reprise 1 - - - - - - - - - ");
		System.out.println("- - - - - - - - - - - - - - - - - - - - - - - - ");
		
		traceln(LOG_PRIORITY_INFO, "launcher executing in " + CUtils.getCurrentLocation() + "...");
		
		// check the configuration sent by the manager
		printConfiguration();

		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		try {
			// retrieve the parameters we need
			restBaseUrl  = getStringParamValue("Mr Suricate", "Mr Suricate URL");
			restEmail    = getStringParamValue("Mr Suricate", "Email");
			restPassword = getStringParamValue("Mr Suricate", "Password");
			customerId   = getStringParamValue("Mr Suricate", "Customer Id");
			intCustomerId = Integer.parseInt(customerId);
			
			browser      = getStringParamValue("Mr Suricate", "Browser");
			
		} catch (CParamParsingException e) {
			traceln(LOG_PRIORITY_SEVERE, "parsing error during initialization");
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Exception during initialize: " + e.getMessage()));
			
			CUtils.showErrorMessage(null, 
					"Error: Outdated configuration", 
					"At least one launcher is requiring some missing configuration parameters.<br/>This may happen when a launcher is updated and needs new or modified parameters.", 
					"Please go to the <b>Configuration</b> tab and create a new configuration for the indicated category<br/>(and of course select this new configuration)",
					JOptionPane.ERROR_MESSAGE);
			
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
		
		// Settings up the connection to the REST API server
		CLauncherRestConnectionManager connectionManager = new CLauncherRestConnectionManager(restBaseUrl, restEmail, restPassword);
		restBackend = new CLauncherRestBackend(connectionManager); // that's where authenticate() is called and the bearer token stored
		
		// ----------------------------------------------------------------------
		// fetch the ids of all the scenario in the workspace (of the customerId)
		// ----------------------------------------------------------------------
		// api-v2/customers/<customerId>/scenarios/get
		String allScenarioJsonContent = restBackend.getAllScenarios(intCustomerId);
		traceln(LOG_PRIORITY_INFO, "- - - - allScenarioJsonContent = " + allScenarioJsonContent);
		JSONObject allScenarioJsonObject = new JSONObject(allScenarioJsonContent);
		JSONArray scenarioJsonArray = allScenarioJsonObject.getJSONArray("items");
		
		for (int i=0; i<scenarioJsonArray.length(); i++) {
			JSONObject scenarioJsonObject = scenarioJsonArray.getJSONObject(i);
			long scenarioId         = scenarioJsonObject.getLong("_id");
			String scenarioName     = scenarioJsonObject.getString("name");
			boolean scenarioEnabled = scenarioJsonObject.getBoolean("enabled");
			//traceln(LOG_PRIORITY_INFO, "- - - - New scenario: scenarioId      = " + scenarioId);
			//traceln(LOG_PRIORITY_INFO, "- - - -               scenarioName    = " + scenarioName);
			//traceln(LOG_PRIORITY_INFO, "- - - -               scenarioEnabled = " + scenarioEnabled);
			
			if (scenarioEnabled)
				hashtableScenarioNameId.put(scenarioName, scenarioId);
		}
		traceln(LOG_PRIORITY_INFO, "- - - - hashtableScenarioNameId = " + hashtableScenarioNameId);
		
		
		// ----------------------------------------------------------------------
		// convert the test name to scenario ids
		// ----------------------------------------------------------------------
		Vector<Long> scenarioToBeExecutedIds = new Vector<Long>();
		for (CTestExecution testExecution : testExecutionVector) {
			String testName = testExecution.getName();
			
			long scenarioToBeExecutedId = -1;
			try {
				scenarioToBeExecutedId = hashtableScenarioNameId.get(testName);
				scenarioToBeExecutedIds.add(scenarioToBeExecutedId);
			} catch (Exception e) {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Couldn't find scenario with name " + testName + "!"));
			}
		}
		traceln(LOG_PRIORITY_INFO, "- - - - scenarioToBeExecutedIds = " + scenarioToBeExecutedIds);
		
		// ----------------------------------------------------------------------
		// Create a schedule that will be executed immediately
		// ----------------------------------------------------------------------
		// api-v2/customers/<customerId>/runs/schedule
		String runScheduleJsonContent = restBackend.runSchedule(intCustomerId, scenarioToBeExecutedIds);
		traceln(LOG_PRIORITY_INFO, "- - - - ALL runIDS = " + runScheduleJsonContent);
		JSONArray allRunScheduleJsonObject = new JSONArray(runScheduleJsonContent);
		for (int i = 0; i < allRunScheduleJsonObject.length(); i++) {
            JSONArray runScheduleJsonArray = allRunScheduleJsonObject.getJSONArray(i);
            for (int j = 0; j < runScheduleJsonArray.length(); j++) {
                JSONObject runObj= runScheduleJsonArray.getJSONObject(j);
                long runId      = runObj.getLong("_id");
                long scenarioId = runObj.getLong("scenario");
    			//traceln(LOG_PRIORITY_INFO, "- - - - New Run: scenarioId = " + scenarioId);
    			//traceln(LOG_PRIORITY_INFO, "- - - -          runId      = " + runId);
    			hashtableScenarioRunIds.put(scenarioId, runId);
            }
        }
		traceln(LOG_PRIORITY_INFO, "- - - - runIds = " + hashtableScenarioRunIds);

		
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus preRun(int testId, String testPath, String testName, Vector<CTestAttribute> attributes, String additionalInfo) {
		traceln(LOG_PRIORITY_INFO, "preRun testId=" + testId + " testPath=" + testPath + " [" + testName + "]...");
		return new CReturnStatus(RESULT_SUCCESS);
	}

	@Override
	public CReturnStatus run(int testId, String testPath, String testName, int testcaseId, int testcaseIndex, String testcaseName, Vector<CTestcaseParam> params, String additionalInfo) {
		traceln(LOG_PRIORITY_INFO, "run testId=" + testId + " testPath=" + testPath + "/" + testName + " testcaseIndex=" + testcaseIndex + "...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();

		// testName --> scenarioId
		//System.out.println("- - - - - hashtableScenarioNameId = " + hashtableScenarioNameId);
		//System.out.println("- - - - - testName                = " + testName);
		long scenarioId = -1;
		try {
			scenarioId = hashtableScenarioNameId.get(testName);
		} catch (Exception e) {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Couldn't find scenario with name " + testName + "!"));
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
		
		// sceanrioId --> runId
		//System.out.println("- - - - - hashtableScenarioRunIds = " + hashtableScenarioRunIds);
		//System.out.println("- - - - - scenarioId              = " + scenarioId);
		long runId = -1;
		try {
			runId = hashtableScenarioRunIds.get(scenarioId);
		} catch (Exception e) {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Couldn't find a run for sceanrio " + scenarioId + "!"));
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
		
		// ----------------------------
		// wait for the completion
		// ----------------------------
		String jsonContent = null;
		JSONObject jsonObj = null;
		for (int i=0; i< 60*60; i++) { // 1 hour max
			// api-v2/customers/<customerId>/runs/<run-id>/get
			jsonContent = restBackend.getRunStatus(intCustomerId, runId);
			//System.out.println("- - - - - jsonContent = " + jsonContent);
			
			jsonObj = new JSONObject(jsonContent);
			String status = jsonObj.getString("status");
			traceln(LOG_PRIORITY_INFO, "- - - - status: " + status); // running, receivedResults, processingResults, finalizing, completed
			if (status.equalsIgnoreCase("completed"))
				break;
				
			CUtils.sleep(1000); // 2 seconds
		}
				
		// ----------------------------
		// get the results and traces
		// ----------------------------
		
		int timeConsumption           = jsonObj.getInt("timeConsumption");
		String finishedAt             = jsonObj.getString("finishedAt");
		boolean integrityCheckSuccess = jsonObj.getBoolean("integrityCheckSuccess");
		
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "Time consumption: " + timeConsumption + " milliseconds"));
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "Finished At: " + finishedAt));
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "Integrity Check Success: " + integrityCheckSuccess));
		
		JSONObject outputObj = jsonObj.getJSONObject("output");
		
		
		JSONObject stepsObj = outputObj.getJSONObject("steps");
		Iterator<String> keys = stepsObj.keys();
		Vector<String> stepsDescriptions = new Vector<String>();
        while (keys.hasNext()) {
            String key = keys.next(); // key = "1-1"
            JSONObject stepProperties = stepsObj.getJSONObject(key);
            System.out.println("-------------------- " + key + " ---------------------");
            String stepLabel = stepProperties.getString("label");
            stepsDescriptions.add(key + ": [b]" + stepLabel + "[/b]");
        }
        
        // Reorder the steps - json.org does not preserve the order of the properties when iterating
        CUtils.sortStringVector(stepsDescriptions);
        
        // and include the steps in the log
        for (String stepDescription: stepsDescriptions) {
        	executionSteps.add(new CExecutionStep(RESULT_INFO, stepDescription));
        }
        
		JSONObject conclusionObj = outputObj.getJSONObject("conclusion");
		
		int isSuccess  = conclusionObj.getInt("isSuccess");
		String lastMessage = conclusionObj.getString("lastMessage");

		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), isSuccess == 1 ? RESULT_SUCCESS : RESULT_FAILURE, lastMessage));
		
		
		// ----------------------------
		// get the end screen capture
		// ----------------------------
		// api-v2/customers/<customerId>/runs/<runid>/screenshots/post
		jsonContent = restBackend.getLastScreenshot(intCustomerId, runId);
		System.out.println("- - - - - jsonContent = " + jsonContent);
		
		try {
			jsonObj = new JSONObject(jsonContent);

			String fileName    = jsonObj.getString("failureName");
			String fileContent = jsonObj.getString("failureFile");
			traceln(LOG_PRIORITY_INFO, "- - - - fileName   : " + fileName);
			traceln(LOG_PRIORITY_INFO, "- - - - fileContent: " + fileContent);
			
			File screenshotFile = CFileUtils.createTmpFile(fileName);
			CFileUtils.deleteAndTouchFile(screenshotFile);
			
			CImageUtils.base64ToImage(fileContent, screenshotFile);
			
			this.addTestcaseMessageAttachment(screenshotFile);
		} catch (JSONException e) {
			System.out.println("- - - - - no failure screen capture!");
		}

		
		return new CReturnStatus(isSuccess == 1 ? RESULT_SUCCESS : RESULT_FAILURE, executionSteps);
	}

	@Override
	public CReturnStatus postRun(int testId, String testPath, String testName) {
		traceln(LOG_PRIORITY_INFO, "postRun testId=" + testId + " testPath=" + testPath + ":" + testName + "...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "postRun: succeeded"));
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus terminate() {
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "Terminate"));
		
		restBackend = null;
		
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}


	// +---------------------------------------------+
	// |               SCANNER FEATURE               |
	// +---------------------------------------------+
	
	@Override
	public boolean canScanPhysicalTests() {
		return false;
	}
	
	@Override
	public Vector<CScannedTest> scanPhysicalTests() {
		Vector<CScannedTest> output = new Vector<CScannedTest>();
		
		/*
		File testRootPathFolder = new File(CFileUtils.getFilePathWithValidSeparator(testRootPath)); // OS dependant path
		if (testRootPathFolder == null || !testRootPathFolder.exists() || !testRootPathFolder.isDirectory()) {
			traceln(LOG_PRIORITY_SEVERE, "Test root path does not exist or is not a directory!");
			return output;
		}

		recursivelyScanFolderForTests(output,
							  testRootPathFolder,      // folder to scan
							  "^.*$",                  // any folder
							  "^.*\\x2etxt$",          // only *.txt files
							  true,                    // remove extension from the file name to determine the test name
							  "/",                     // separator used in canonicalPath generated
							  "");                     // canonical path of the folders to start to scan (from testRootPathFolder)
		*/
		
		return output;
	}
}
