/*
+----------------------------------------------------------------------+
|                    Class: CLauncher                                  |
|                                                                      |
| Developper:  Eric Gavaldo (XQual)                                    |
| Version:     1.0                                                     |
| License:     GPL v2 (https://www.gnu.org/licenses/gpl-2.0.html)      |
+----------------------------------------------------------------------+
*/

package com.xqual.xlauncher.playwright;

import java.io.File;
import java.util.Vector;

import javax.swing.JOptionPane;

import org.json.JSONArray;
import org.json.JSONObject;

import com.xqual.xagent.CScannedTest;
import com.xqual.xagent.launcher.CExecutionStep;
import com.xqual.xagent.launcher.CLauncher;
import com.xqual.xagent.launcher.CReturnStatus;
import com.xqual.xagent.launcher.runner.CRunner;
import com.xqual.xagent.launcher.runner.IRunner;
import com.xqual.xcommon.CNetworkInfo;
import com.xqual.xcommon.IConstantsResults;
import com.xqual.xcommon.type.CParamParsingException;
import com.xqual.xcommon.type.CTestAttribute;
import com.xqual.xcommon.type.CTestcaseParam;
import com.xqual.xcommon.utils.CCalendarUtils;
import com.xqual.xcommon.utils.CFileUtils;
import com.xqual.xcommon.utils.CUtils;


/**
 * The <code>CLauncherImpl</code> implementation of <code>ILauncher</code> for ant build files.
 * @author egavaldo
 */
public class CLauncherImpl extends CLauncher implements IConstantsResults {

	// +==============================================================+
	// | Attributes                                                   |
	// +==============================================================+

	static final String TRACE_HEADER = "{playwright  }  ";

	static String NPX_EXECUTABLE = "";
	
	// parameters impacting executing at run time set by the test operator
	private String testRootPath;
	
	private String language;
	private String nodeJsInstallPath;
	private String commandOptions;

	private File workingDir;
	
	private File reportJSONFile;
	
	private String fileExtension;
	
	// +==============================================================+
	// | Constructors                                                 |
	// +==============================================================+

	public CLauncherImpl() {
		super(TRACE_HEADER);
	}

	// +==============================================================+
	// | Methods                                                      |
	// +==============================================================+

	@Override
	public CReturnStatus initialize(int sutId, String sutName, String sutVersion) { 
		setSutDetails(sutId, sutName, sutVersion);

		System.out.println("------------------------------");
		System.out.println("Saturday 17:17");
		System.out.println("------------------------------");
		
		traceln(LOG_PRIORITY_INFO, "launcher executing in " + CUtils.getCurrentLocation() + "...");
		
		// check the configuration sent by the manager
		printConfiguration();

		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		try {
			// retrieve the parameters we need
			testRootPath    	= getStringParamValue("General",  "Test root path");
			
			language			= getStringParamValue("Playwright",  "Language");
			nodeJsInstallPath	= getStringParamValue("Playwright",  "Node.js install path");
			commandOptions		= getStringParamValue("Playwright",  "Command options");
			
			if (language.equalsIgnoreCase("TypeScript")) {
				fileExtension = ".ts";
			} else {
				fileExtension = ".js";
			}
			
		} catch (CParamParsingException e) {
			traceln(LOG_PRIORITY_SEVERE, "parsing error during initialization");
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Exception during initialize: " + e.getMessage()));
			
			CUtils.showErrorMessage(null, 
					"Error: Outdated configuration", 
					"At least one launcher is requiring some missing configuration parameters.<br/>This may happen when a launcher is updated and needs new or modified parameters.", 
					"Please go to the <b>Configuration</b> tab and create a new configuration for the indicated category<br/>(and of course select this new configuration)",
					JOptionPane.ERROR_MESSAGE);
			
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
		
		// +------------------------------------+
		// | Set environment variables
		// +------------------------------------+
		
    	CNetworkInfo networkInfo = new CNetworkInfo();
    	short osType = networkInfo.getOsType();
    	
		switch (osType) {
			case CNetworkInfo.OS_WIN32: 
			case CNetworkInfo.OS_WIN64: 
				NPX_EXECUTABLE = CFileUtils.quoteFilePath(nodeJsInstallPath + "\\npx.cmd") + " playwright test";
    			break;
    		case CNetworkInfo.OS_LINUX: 
    		case CNetworkInfo.OS_MAC:
    			NPX_EXECUTABLE = CFileUtils.quoteFilePath(nodeJsInstallPath + "\\npx") + " playwright test";
    			break;
    		case CNetworkInfo.OS_UNKNOWN:
    		default:
    			traceln(LOG_PRIORITY_SEVERE, "OS couldn't be determined!");
    			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "OS couldn't be determined!"));
    			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
		
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus preRun(int testId, String testPath, String testName, Vector<CTestAttribute> attributes, String additionalInfo) {
		traceln(LOG_PRIORITY_INFO, "preRun testId=" + testId + " testPath=" + testPath + " [" + testName + "]...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus run(int testId, String testPath, String testName, int testcaseId, int testcaseIndex, String testcaseName, Vector<CTestcaseParam> params, String additionalInfo) {
		traceln(LOG_PRIORITY_INFO, "run testId=" + testId + " testPath=" + testRootPath + "/" + testPath + "/" + testName + " testcaseIndex=" + testcaseIndex + "...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();

		workingDir = new File(testRootPath);
		
		reportJSONFile = new File (workingDir + "/jsonReport.json");
		CFileUtils.deleteAndTouchFile(reportJSONFile);
		
		String fullTestPath = (testPath.trim().length() > 0 ? testPath + "/" + testName : testName);
		
		String playwrightInterpreter = NPX_EXECUTABLE + " " + CFileUtils.quoteFilePath(fullTestPath) + " ";											// run the test (e.g. "npx.cmd playwright test folder1/example")
		playwrightInterpreter = playwrightInterpreter + "--reporter=json > " + CFileUtils.quoteFilePath(reportJSONFile.getAbsolutePath()) + " ";	// generate a json report
		playwrightInterpreter = playwrightInterpreter + commandOptions;																				// add the optional custom options
		
		
		// +------------------------------------+
		// | Interpret the script
		// +------------------------------------+
		CRunner playwrightRunner = new CRunner("[" + testId + "] "+ testPath + ":" + testName + "." + testcaseIndex,
											   playwrightInterpreter,
											   /*environment,*/
											   workingDir);
		
		short result = playwrightRunner.requestAction(IRunner.START_PROCESS, IRunner.WAIT_END_OF_EXECUTION);
		traceln(LOG_PRIORITY_INFO, "returned from playwright runner: " + result);

		if (result == RESULT_FAILURE) {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "execution returned a code > 0"));
		} else {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_SUCCESS, "execution succeeded"));
		}
		
		if (!reportJSONFile.exists()) {
			traceln(LOG_PRIORITY_SEVERE, "JSON report file not found!");
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "run: JSON report file not found!"));
			return new CReturnStatus(RESULT_FAILURE, executionSteps); // nothing to parse
		} else {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_SUCCESS, "JSON report file exists"));
			if (reportJSONFile.length() > 0) {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_SUCCESS, "JSON report file not empty"));
			} else {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "JSON report file is empty!"));
				return new CReturnStatus(RESULT_FAILURE, executionSteps); // nothing to parse
			}
			
			addTestcaseMessageAttachment(reportJSONFile);
		}


		
		return parseResultFile(executionSteps);
	}

	@Override
	public CReturnStatus postRun(int testId, String testPath, String testName) {
		traceln(LOG_PRIORITY_INFO, "postRun testId=" + testId + " testPath=" + testPath + ":" + testName + "...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_SUCCESS, "postRun: succeeded"));
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus terminate() {
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_SUCCESS, "Terminate"));
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	// +--------------------------+
	// | Utilities                |
	// +--------------------------+

	private CReturnStatus parseResultFile(Vector<CExecutionStep> executionSteps) {
		traceln(LOG_PRIORITY_INFO, "-------------------------------------");
		traceln(LOG_PRIORITY_INFO, "Parsing the json report...");
		traceln(LOG_PRIORITY_INFO, "-------------------------------------");

		Boolean[] errorDetected = {false};
		
		// parse the json report file to get the details so we can add useful information in the message list
		//System.err.println("- - - - json file: " + CFileUtils.FileToString(reportJSONFile));

		try {
			JSONObject jsonRootObj= new JSONObject(CFileUtils.FileToString(reportJSONFile));
			
			JSONArray suiteArray = jsonRootObj.getJSONArray("suites");
			for (int i=0; i<suiteArray.length(); i++) {
				System.out.println("- - - suite " + i + "\n");
				JSONObject suiteObj = suiteArray.getJSONObject(i);
				
				String suiteTitle = suiteObj.getString("title"); // e.g. "subfolder\\demo-todo-app.spec.ts"
				//executionSteps.add(new CExecutionStep(RESULT_INFO, "SUITE [b]" + suiteTitle.toUpperCase() + "[/b]"));

				System.out.println("********** call parseInnerSpecs");
				parseInnerSpecs(suiteObj, executionSteps, errorDetected);
			}
		} catch (Exception e) {
			traceln(LOG_PRIORITY_SEVERE, "Couldn't parse the json: " + e);
			executionSteps.add(new CExecutionStep(RESULT_FAILURE, "Couldn't parse the json"));
			errorDetected[0]=true;
		}
		
		if (errorDetected[0]==true) {
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		} else {
			return new CReturnStatus(RESULT_SUCCESS, executionSteps);
		}
	}
	
	// must be recursive
	private void parseInnerSpecs(JSONObject suiteObj, Vector<CExecutionStep> executionSteps, Boolean[] errorDetected) {
		System.out.println("********** parseInnerSpecs");
		System.out.println("********** suiteObj = " + suiteObj.toString());
		System.out.println("********** executionSteps = " + executionSteps.toString());
		
		try {
			String previousBrowser = "";
			JSONArray specArray = suiteObj.getJSONArray("specs");
			System.out.println("********** nbSpecs " + specArray.length());
			for (int j=0; j<specArray.length(); j++) {
				System.out.println("- spec " + j);
				JSONObject specObj = specArray.getJSONObject(j);
				
				String specTitle = specObj.getString("title");
				
				JSONArray testArray = specObj.getJSONArray("tests");
				for (int k=0; k<testArray.length(); k++) {
					System.out.println("- - test " + k);
					JSONObject testObj = testArray.getJSONObject(k);
					
					String browser = testObj.getString("projectName");
					if (!browser.equalsIgnoreCase(previousBrowser)) {
						executionSteps.add(new CExecutionStep(RESULT_INFO, "[" + browser.toUpperCase() + "]"));
					}
					previousBrowser = browser;
					
					JSONArray resultsArray = testObj.getJSONArray("results");
					for (int m=0; m<resultsArray.length(); m++) {
						System.out.println("- - - result " + m);
						JSONObject resultObj = resultsArray.getJSONObject(m);
					
						String status    = resultObj.getString("status");
						long duration    = resultObj.getLong("duration");
						String timestamp = resultObj.getString("startTime");
						
						boolean passed = status.equalsIgnoreCase("passed");
						if (!passed) errorDetected[0]=true;
						
						executionSteps.add(new CExecutionStep(timestamp, (passed ? RESULT_SUCCESS : RESULT_FAILURE), "[" + m + "] [b]" + specTitle + "[/b]: " + status + " (duration: " + CCalendarUtils.millisTimeToDurationString(duration) + ")"));
					}
				}
			}
			
			// recursive call... check if there are sub suites...
			try {
				JSONArray subSuiteArray = suiteObj.getJSONArray("suites");
				for (int j=0; j<subSuiteArray.length(); j++) {
					System.out.println("- - - - - - subsuite " + j + "\n");
					JSONObject subSuiteObj = subSuiteArray.getJSONObject(j);
					
					String subSuiteTitle = subSuiteObj.getString("title"); // e.g. "New Todo"
					executionSteps.add(new CExecutionStep(RESULT_INFO, "> SUITE [b]" + subSuiteTitle.toUpperCase() + "[/b]"));
					
					System.out.println("********** call parseInnerSpecs");
					parseInnerSpecs(subSuiteObj, executionSteps, errorDetected);
				}
			} catch (Exception e) {
				System.out.println("- - - - - - NO subsuite\n");
			}
		} catch (Exception e) {
			traceln(LOG_PRIORITY_SEVERE, "Couldn't parse the sub-json: " + e);
			executionSteps.add(new CExecutionStep(RESULT_FAILURE, "Couldn't parse the sub-json"));
			errorDetected[0]=true;
		}
	}
	
	
	// +---------------------------------------------+
	// |               SCANNER FEATURE               |
	// +---------------------------------------------+
	
	@Override
	public boolean canScanPhysicalTests() {
		return true;
	}
	
	@Override
	public Vector<CScannedTest> scanPhysicalTests() {
		Vector<CScannedTest> output = new Vector<CScannedTest>();
		Vector<CScannedTest> playwrightOutput = new Vector<CScannedTest>();
		String playwrightCanonicalPath;
		String playwrightTestName;		
		
		File testRootPathFolder = new File(CFileUtils.getFilePathWithValidSeparator(testRootPath)); // OS dependplaywright path
		if (testRootPathFolder == null || !testRootPathFolder.exists() || !testRootPathFolder.isDirectory()) {
			traceln(LOG_PRIORITY_SEVERE, "Test root path does not exist or is not a directory!");
			return output;
		}

		recursivelyScanFolderForTests(output,
							  testRootPathFolder,      // folder to scan
							  "^.*$",                  // any folder
							  "^.*\\x2ets$",           // only *.ts or *.js files
							  true,                    // remove extension from the file name to determine the test name
							  "/",                     // separator used in canonicalPath generated
							  "");                     // canonical path of the folders to start to scan (from testRootPathFolder)
		
			
		for (CScannedTest scannedTest : output) {
			playwrightCanonicalPath = scannedTest.getCanonicalPath();
			playwrightTestName      = scannedTest.getName();
			playwrightOutput.add(new CScannedTest(playwrightTestName, playwrightCanonicalPath));
		}

		return playwrightOutput;
	}

}
