/*
+----------------------------------------------------------------------+
|                    Class: CLauncher                                  |
|                                                                      |
| Developper:  Eric Gavaldo (eric.gavaldo@xqual.com)                   |
| Version:     1.8                                                     |
| License:     GPL v2 (https://www.gnu.org/licenses/gpl-2.0.html)      |
+----------------------------------------------------------------------+
*/
package com.xqual.xlauncher.vtaas;

import java.util.Vector;

import javax.swing.JOptionPane;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.xqual.xagent.CScannedTest;
import com.xqual.xagent.launcher.CExecutionStep;
import com.xqual.xagent.launcher.CLauncher;
import com.xqual.xagent.launcher.CReturnStatus;
import com.xqual.xcommon.CXmlDocumentFactory;
import com.xqual.xcommon.IConstantsResults;
import com.xqual.xcommon.type.CParamParsingException;
import com.xqual.xcommon.type.CTestAttribute;
import com.xqual.xcommon.type.CTestcaseParam;
import com.xqual.xcommon.utils.CCalendarUtils;
import com.xqual.xcommon.utils.CConversionUtils;
import com.xqual.xcommon.utils.CTextUtils;
import com.xqual.xcommon.utils.CUtils;



/**
 * The <code>CLauncherImpl</code> implementation of <code>ILauncher</code> for Cerberus.
 * @author egavaldo
 */
public class CLauncherImpl extends CLauncher implements IConstantsResults {

	// +==============================================================+
	// | Attributes                                                   |
	// +==============================================================+

	static final String TRACE_HEADER = "{vtaas         }  ";

	private String restBaseUrl;
	private String restApiKey;

	private String targetBaseUrl;

	private CLauncherRestBackend restBackend;
	
	// +==============================================================+
	// | Constructors                                                 |
	// +==============================================================+

	public CLauncherImpl() {
		super(TRACE_HEADER);
	}

	// +==============================================================+
	// | Methods                                                      |
	// +==============================================================+

	@Override
	public CReturnStatus initialize(int sutId, String sutName, String sutVersion) { 
		setSutDetails(sutId, sutName, sutVersion);

		System.out.println("- - - - - - - - - - - - - - - - - - - - - - - - ");
		System.out.println("- - - - - - - - - - - V2 - - - - - - - - - - -  ");
		System.out.println("- - - - - - - - - - - - - - - - - - - - - - - - ");
		
		traceln(LOG_PRIORITY_INFO, "launcher executing in " + CUtils.getCurrentLocation() + "...");
		
		// check the configuration sent by the manager
		printConfiguration();

		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		try {
			// retrieve the parameters we need
			restBaseUrl = getStringParamValue("REST Connection", "Base URL");
			restApiKey  = getStringParamValue("REST Connection", "API Key");
			
			targetBaseUrl = getStringParamValue("Target", "Base URL");
		} catch (CParamParsingException e) {
			traceln(LOG_PRIORITY_SEVERE, "parsing error during initialization");
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Exception during initialize: " + e.getMessage()));
			
			CUtils.showErrorMessage(null, 
					"Error: Outdated configuration", 
					"At least one launcher is requiring some missing configuration parameters.<br/>This may happen when a launcher is updated and needs new or modified parameters.", 
					"Please go to the <b>Configuration</b> tab and create a new configuration for the indicated category<br/>(and of course select this new configuration)",
					JOptionPane.ERROR_MESSAGE);
			
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
		
		// Settings up the connection to the REST API server
		CLauncherRestConnectionManager connectionManager = new CLauncherRestConnectionManager(restBaseUrl, restApiKey);
		restBackend = new CLauncherRestBackend(connectionManager);
		
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus preRun(int testId, String testPath, String testName, Vector<CTestAttribute> attributes, String additionalInfo) {
		traceln(LOG_PRIORITY_INFO, "preRun testId=" + testId + " testPath=" + testPath + " [" + testName + "]...");
		return new CReturnStatus(RESULT_SUCCESS);
	}

	@Override
	public CReturnStatus run(int testId, String testPath, String testName, int testcaseId, int testcaseIndex, String testcaseName, Vector<CTestcaseParam> params, String additionalInfo) {
		traceln(LOG_PRIORITY_INFO, "run testId=" + testId + " testPath=" + testPath + "/" + testName + " testcaseIndex=" + testcaseIndex + "...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();

		/*
		We must pass a json body similar to:
		{
		  "description": "string",
		  "url": "string",
		  "scenario": {
		    "name": "Expérimenté-Paris-MALUS-Assurance tous risques-TARIF",
		    "actions": [
		      {
		        "action": "Recharger la page courante",
		        "expectedResult": ""
		      },
		      {
		        "action": "Rechercher un véhicule ayant les caractèristiques suivantes :",
		        "expectedResult": ""
		      },
		      {
		        "action": "Cliquer sur simuler",
		        "expectedResult": "Vérifier l'existence du texte Tarif annuel : 4155.84"
		      }
		    ]
		  }
		}
		*/
		
		StringBuffer jsonProcedureBuffer = new StringBuffer();
		jsonProcedureBuffer.append("{");
		jsonProcedureBuffer.append("\"description\": \"Execution from XQual\",");
		jsonProcedureBuffer.append("\"url\": \"" + targetBaseUrl + "\",");
		jsonProcedureBuffer.append("\"scenario\": {");
		
		jsonProcedureBuffer.append("\"name\": \"" + CTextUtils.escapeJsonValue(testcaseName) + "\",");
		jsonProcedureBuffer.append("\"actions\": [");

		String xmlContent = getTestcaseStepsLinkedHashMap().get(testcaseId);
    	traceln(LOG_PRIORITY_INFO, "xmlContent = " + xmlContent);
    	Document xmlDocument = CXmlDocumentFactory.createXMLDoc(xmlContent);
	    NodeList stepNodeList = CXmlDocumentFactory.getNodeObjListFromXPath(xmlDocument, "//object[@type='step']");
	    
	    // for each step...
	    for (int j=0; j<stepNodeList.getLength(); j++) {
	    	jsonProcedureBuffer.append("{");
	    	
			Node stepNode = stepNodeList.item(j);
	
	    	// Step
	    	String stepDescription = CTextUtils.unescapeAttributeString(CXmlDocumentFactory.getNodeStringValueFromXPath(stepNode, "@description"));
	    	stepDescription = stepDescription.trim();
	    	String plainTextStepDescription = CConversionUtils.pseudoHtmlToPlainTextFormatting(stepDescription);
	    	jsonProcedureBuffer.append("\"action\": \"" + CTextUtils.escapeJsonValue(plainTextStepDescription) + "\",");
	    	
	    	// checks
	    	StringBuffer checkListBuffer = new StringBuffer();
			NodeList checkNodeList = CXmlDocumentFactory.getNodeObjListFromXPath(stepNode, ".//object[@type='check']");
			for (int k=0; k<checkNodeList.getLength(); k++) {
				Node checkNode = checkNodeList.item(k);
				
				String checkDescription = CTextUtils.unescapeAttributeString(CXmlDocumentFactory.getNodeStringValueFromXPath(checkNode, "@description"));
				checkDescription = checkDescription.trim();
				String plainTextCheckDescription = CConversionUtils.pseudoHtmlToPlainTextFormatting(checkDescription);
				
				checkListBuffer.append(plainTextCheckDescription);
				if (k < (checkNodeList.getLength()-1))
					checkListBuffer.append(", ");
			}
			jsonProcedureBuffer.append("\"expectedResult\": \"" + CTextUtils.escapeJsonValue(checkListBuffer.toString()) + "\"");
			
			jsonProcedureBuffer.append("}");
			if (j < (stepNodeList.getLength()-1))
				jsonProcedureBuffer.append(",");
	    }
		jsonProcedureBuffer.append("]");
		
		jsonProcedureBuffer.append("}");
		jsonProcedureBuffer.append("}");
		
		// +-----------------+
		// |       RUN       | 
		// +-----------------+
		System.out.println("- - - - running testcase " + testcaseId + " : " + testcaseName);
		System.out.println("- - - - procedure in vtaas format " + jsonProcedureBuffer.toString());
		
		String runResponse = restBackend.run(jsonProcedureBuffer.toString());
		System.out.println("- - - - runResponse " + runResponse);
		if (runResponse == null) {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Not a 200 or 201 status code, check your configuration!"));
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		}
			
		JSONObject runResponseRootObj = new JSONObject(runResponse);
		/*
		{
		   "status": {
		      "id": 103,
		      "timestamp": 1737927584119,
		      "state": "running",  // running, failed, terminated
		      "doneSteps": [
		         {"name":"xxxxx", "timestamp": xxxxx},
		         {"name":"xxxxx", "timestamp": xxxxx}] 
		      "verdict": {											<-- only when state is 'failed' or 'terminated'
		         "result": "pass"      // pass, fail, error
		      }
		   }
		}
		
		or
		
		{
		  "error": "xxx"
		}
		*/
		
		int testcaseRunId = -1;
		String testcaseReason = "";
		
		// status
		try {
			testcaseRunId = runResponseRootObj.getInt("id");
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "TC_" + testcaseId + " started..."));
		} catch (JSONException e) {
			try {
				String errorMessage = runResponseRootObj.getString("error");
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Error while starting TC_" + testcaseId + ": " + errorMessage));
				return new CReturnStatus(RESULT_FAILURE, executionSteps);
			} catch (JSONException e1) {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Error while starting TC_" + testcaseId + "!"));
				return new CReturnStatus(RESULT_FAILURE, executionSteps);
			}
		}

		System.out.println("=========== WAIT =============");
		
		CUtils.sleep(5000);
		
		// --------------------------
		//  WAITING END OF EXECUTION
		// --------------------------

		System.out.println("=========== WAITING END OF EXECUTION =============");

		int counter=0;
		String testcaseState = "running";
		long timestamp = 0;
		while (testcaseState.equalsIgnoreCase("running")) {
			counter++;

			System.out.println("====> Wait #" + counter);

			// timeout
			if (counter >= 3600) { // 1 hour
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Timeout while starting TC_" + testcaseId + " (" + counter + " seconds)"));
				return new CReturnStatus(RESULT_FAILURE, executionSteps);
			}
			
			//System.out.println("- - - - requesting status for " + testcaseId + "...");
			String statusResponse = restBackend.getStatus(testcaseRunId);
			System.out.println("- - - - statusResponse " + statusResponse);
			if (statusResponse == null) {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Not a 200 or 201 status code, weird that status return nothing... check VTaaS's log:"));
				return new CReturnStatus(RESULT_FAILURE, executionSteps);
			}
			JSONObject statusResponseRootObj = new JSONObject(statusResponse);
			
			try {
				// +----------------+
				// |      State     |
				// +----------------+
				testcaseState = statusResponseRootObj.getString("state");                     // i.e. "running, failed, terminated"
				timestamp     = statusResponseRootObj.getLong("timestamp");                   // i.e. 1737928612089
				
				
				if (testcaseState.equalsIgnoreCase("failed")) {
					// doneSteps
					addDoneStepsToExecutionSteps(statusResponseRootObj, executionSteps);
					
					executionSteps.add(new CExecutionStep(CCalendarUtils.millisTimeToDateString(timestamp), RESULT_FAILURE, "TC_" + testcaseId + ": FAILURE"));
					return new CReturnStatus(RESULT_FAILURE, executionSteps);
					
					
				} else if (testcaseState.equalsIgnoreCase("terminated")) {                   // TERMINATED !
					// doneSteps
					addDoneStepsToExecutionSteps(statusResponseRootObj, executionSteps);

					// verdict
					JSONObject verdictObj = statusResponseRootObj.getJSONObject("verdict");
					String testcaseResult = verdictObj.getString("result");                  // VERDICT: pass, fail, error
					
					testcaseReason = "";                                                     // REASON
					try {testcaseReason = "\n" + verdictObj.getString("reason");} catch (Exception e) {}
					
					if (testcaseResult.equalsIgnoreCase("fail")) {
						executionSteps.add(new CExecutionStep(CCalendarUtils.millisTimeToDateString(timestamp), RESULT_FAILURE, "TC_" + testcaseId + ": FAILURE" + testcaseReason));
						return new CReturnStatus(RESULT_FAILURE, executionSteps);
						
					} else if (testcaseResult.equalsIgnoreCase("error")) {
						executionSteps.add(new CExecutionStep(CCalendarUtils.millisTimeToDateString(timestamp), RESULT_FAILURE, "TC_" + testcaseId + ": ERROR" + testcaseReason));
						return new CReturnStatus(RESULT_FAILURE, executionSteps);
						
					} else if (testcaseResult.equalsIgnoreCase("pass")) {
						//executionSteps.add(new CExecutionStep(CCalendarUtils.millisTimeToDateString(timestamp), RESULT_SUCCESS, "TC_" + testcaseId + ": SUCCESS" + testcaseReason));
						break;
					}
				}
			} catch (JSONException e) {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "Error while getting status of testcase run " + testcaseRunId + "!"));
				return new CReturnStatus(RESULT_FAILURE, executionSteps);
			}
			
			CUtils.sleep(1000);
		}
		
		System.out.println("=========== FETCHING THE VIDEO =============");

		String videoResponse = restBackend.getVideo(testcaseRunId);
		System.out.println("- - - - videoResponse = " + videoResponse);
		String videoUrl;
		try {
			if (videoResponse == null) {
				executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_FAILURE, "could not fetch the video!"));
			} else {
				JSONObject videoResponseRootObj = new JSONObject(videoResponse);
				videoUrl = videoResponseRootObj.getString("url");
	
				executionSteps.add(new CExecutionStep(RESULT_INFO, "Video of the execution: [a]" + videoUrl + "[/a]"));
			}
		} catch (JSONException e) {
		}

		/*
		if (videoFile == null) {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "No video file available!"));
			return new CReturnStatus(RESULT_FAILURE, executionSteps);
		} else {
			executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "Video file available as an attachment"));
			addTestcaseMessageAttachment(videoFile);
		}
		*/
		executionSteps.add(new CExecutionStep(CCalendarUtils.millisTimeToDateString(timestamp), RESULT_SUCCESS, "TC_" + testcaseId + ": SUCCESS" + testcaseReason));
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	private void addDoneStepsToExecutionSteps(JSONObject statusObj, Vector<CExecutionStep> executionSteps) {
		try {
			JSONArray doneStepsObjArray = statusObj.getJSONArray("doneSteps"); // i.e. "[{"name":"xxxxx", "timestamp": xxxxx}, {"name":"xxxxx", "timestamp": xxxxx}]"

			for (int i=0; i<doneStepsObjArray.length(); i++) {
				JSONObject doneStepsObj = doneStepsObjArray.getJSONObject(i);
				
				String stepName      = doneStepsObj.getString("name");
				long startedAt       = doneStepsObj.getLong("startedAt");
				long finishedAt      = doneStepsObj.getLong("finishedAt");
				String screenshotUrl = doneStepsObj.getString("screenshotUrl");
				String duration = CCalendarUtils.millisTimeToDurationString(finishedAt-startedAt);
				
				executionSteps.add(new CExecutionStep(CCalendarUtils.millisTimeToDateString(startedAt), RESULT_INFO, stepName + " (duration: " + duration + ")"));
				executionSteps.add(new CExecutionStep(RESULT_INFO, "[a]" + screenshotUrl + "[/a]"));
			}
		} catch (Exception e) {
			System.out.println("- - -  couldn't retrieve doneSteps");
		}
	}
	
	
	@Override
	public CReturnStatus postRun(int testId, String testPath, String testName) {
		traceln(LOG_PRIORITY_INFO, "postRun testId=" + testId + " testPath=" + testPath + ":" + testName + "...");
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "postRun: succeeded"));
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}

	@Override
	public CReturnStatus terminate() {
		Vector<CExecutionStep> executionSteps = new Vector<CExecutionStep>();
		executionSteps.add(new CExecutionStep(CCalendarUtils.getCurrentTime(), RESULT_INFO, "Terminate"));
		
		restBackend = null;
		
		return new CReturnStatus(RESULT_SUCCESS, executionSteps);
	}


	// +---------------------------------------------+
	// |               SCANNER FEATURE               |
	// +---------------------------------------------+
	
	@Override
	public boolean canScanPhysicalTests() {
		return false;
	}
	
	@Override
	public Vector<CScannedTest> scanPhysicalTests() {
		Vector<CScannedTest> output = new Vector<CScannedTest>();
		
		/*
		File testRootPathFolder = new File(CFileUtils.getFilePathWithValidSeparator(testRootPath)); // OS dependant path
		if (testRootPathFolder == null || !testRootPathFolder.exists() || !testRootPathFolder.isDirectory()) {
			traceln(LOG_PRIORITY_SEVERE, "Test root path does not exist or is not a directory!");
			return output;
		}

		recursivelyScanFolderForTests(output,
							  testRootPathFolder,      // folder to scan
							  "^.*$",                  // any folder
							  "^.*\\x2etxt$",          // only *.txt files
							  true,                    // remove extension from the file name to determine the test name
							  "/",                     // separator used in canonicalPath generated
							  "");                     // canonical path of the folders to start to scan (from testRootPathFolder)
		*/
		
		return output;
	}
}
